/* $Header: /usr/src/local/etc/tftpd/RCS/tftpd.c,v 1.7 1995/02/20 21:14:41 smb Exp $ */

/*
 * Modified by:	Scott M. Ballew
 * 		Purdue University Computing Center
 * 		Purdue University
 * Date:	Tue Mar 15 13:57:10 1994
 *
 * Modifications Copyright 1994 Purdue Research Foundation,
 * 	West Lafayette, Indiana 47907.  All rights reserved.
 *
 * The modifications in this software are not subject to any license of
 * the American Telephone and Telegraph Company or the Regents of the
 * University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

/*
 * NOTE: This software is a modification of the NET2 tftpd.c sources
 * which are covered by the following copyright notice and use provisions.
 */

/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1983 Regents of the University of California.\n\
 All rights reserved.\n\
 Modifications Copyright 1994 Purdue Research Foundation.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)tftpd.c	5.13 (Berkeley) 2/26/91";
#endif /* not lint */

#ifndef lint
static char rcs_ident[] = "$Id: tftpd.c,v 1.7 1995/02/20 21:14:41 smb Exp $";
#endif

/*
 * $Log: tftpd.c,v $
 * Revision 1.7  1995/02/20  21:14:41  smb
 * Fix a botch in the earlier changes.
 *
 * Revision 1.6  1995/02/20  21:08:41  smb
 * Added Purdue Copyright information.
 *
 */

/*
 * Trivial file transfer protocol server.
 *
 * This version includes many modifications by Jim Guyton <guyton@rand-unix>
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/tftp.h>
#include <netdb.h>

#include <setjmp.h>
#include <syslog.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#ifdef ACCESS_CONTROL
#include <pwd.h>
#include <arpa/inet.h>

#define syslog		UDP_syslog
#define openlog		UDP_openlog
#define closelog	UDP_closelog
#define setlogmask	UDP_setlogmask

#endif

#define	TIMEOUT		5

extern	int errno;
struct	sockaddr_in sin = { AF_INET };
int	peer;
int	rexmtval = TIMEOUT;
int	maxtimeout = 5*TIMEOUT;

#define	PKTSIZE	SEGSIZE+4
char	buf[PKTSIZE];
char	ackbuf[PKTSIZE];
struct	sockaddr_in from;
int	fromlen;

#ifdef ACCESS_CONTROL
int logging = 0;
int restrict = 0;
int secure = 0;
char *homedir = "/tftpboot";	/* the default */

char *host, *haddr;
int access_fail = 0;

struct aentry {
    unsigned long addr;
    unsigned long mask;
    char *pattern;
    struct aentry *next;
} *access_list = NULL;

static struct aentry *read_config();

#ifndef CONFIG_FILE
#define CONFIG_FILE	"/usr/local/etc/tftpd.conf"
#endif

uid_t uid;
gid_t gid;

#define UID_NOBODY	65534
#define GID_NOBODY	65534

#define DELAY_SECS	3	/* amount to delay error on initial conn. */
#else
#define MAXARG	4
char	*dirs[MAXARG+1];
#endif

char *strerror();
extern int sys_nerr;
extern char *sys_errlist[];

main(ac, av)
	char **av;
{
	register struct tftphdr *tp;
	register int n = 0;
	int on = 1;
#ifdef ACCESS_CONTROL
	char c;
	extern int optind;
	struct hostent *hent;
	struct passwd *pw;
#endif

#ifdef ACCESS_CONTROL
#ifdef sun
	tzset();	/* make the syslog stuff happy after chroot() */
#endif
	openlog("tftpd", LOG_PID, LOG_DAEMON);
	while ((c = getopt(ac, av, "lrs")) != -1) {
	    switch (c) {
		case 'l': {
		    logging++;
		    break;
		}
		case 'r': {
		    restrict++;
		    if ((access_list = read_config(CONFIG_FILE)) == NULL) {
			/*
			 * if we have any problems, immediately fail
			 * all access attempts
			 */
			access_fail++;
		    }
		    break;
		}
		case 's': {
		    secure++;
		    break;
		}
		default: {
		    usage(av[0]);
		    break;
		}
	    }
	}
	if (optind < ac) {
	    if ((optind == ac - 1) && (*av[optind] == '/')) {
		homedir = av[optind];
	    }
	    else {
		usage(av[0]);
	    }
	}
	    
#else
	ac--; av++;
	while (ac-- > 0 && n < MAXARG)
		dirs[n++] = *av++;
	openlog("tftpd", LOG_PID, LOG_DAEMON);
#endif
	if (ioctl(0, FIONBIO, &on) < 0) {
		syslog(LOG_ERR, "ioctl(FIONBIO): %m\n");
		exit(1);
	}
	fromlen = sizeof (from);
	n = recvfrom(0, buf, sizeof (buf), 0,
	    (struct sockaddr *)&from, &fromlen);
	if (n < 0) {
		syslog(LOG_ERR, "recvfrom: %m\n");
		exit(1);
	}
	
	/*
	 * Now that we have read the message out of the UDP
	 * socket, we fork and exit.  Thus, inetd will go back
	 * to listening to the tftp port, and the next request
	 * to come in will start up a new instance of tftpd.
	 *
	 * We do this so that inetd can run tftpd in "wait" mode.
	 * The problem with tftpd running in "nowait" mode is that
	 * inetd may get one or more successful "selects" on the
	 * tftp port before we do our receive, so more than one
	 * instance of tftpd may be started up.  Worse, if tftpd
	 * break before doing the above "recvfrom", inetd would
	 * spawn endless instances, clogging the system.
	 */
	{
		int pid;
		int i, j;

		for (i = 1; i < 20; i++) {
		    pid = fork();
		    if (pid < 0) {
				sleep(i);
				/*
				 * flush out to most recently sent request.
				 *
				 * This may drop some request, but those
				 * will be resent by the clients when
				 * they timeout.  The positive effect of
				 * this flush is to (try to) prevent more
				 * than one tftpd being started up to service
				 * a single request from a single client.
				 */
				j = sizeof from;
				i = recvfrom(0, buf, sizeof (buf), 0,
				    (struct sockaddr *)&from, &j);
				if (i > 0) {
					n = i;
					fromlen = j;
				}
		    } else {
				break;
		    }
		}
		if (pid < 0) {
			syslog(LOG_ERR, "fork: %m\n");
			exit(1);
		} else if (pid != 0) {
			exit(0);
		}
	}
	from.sin_family = AF_INET;
	alarm(0);
	close(0);
	close(1);
	peer = socket(AF_INET, SOCK_DGRAM, 0);
	if (peer < 0) {
		syslog(LOG_ERR, "socket: %m\n");
		exit(1);
	}
	if (bind(peer, (struct sockaddr *)&sin, sizeof (sin)) < 0) {
		syslog(LOG_ERR, "bind: %m\n");
		exit(1);
	}
	if (connect(peer, (struct sockaddr *)&from, sizeof(from)) < 0) {
		syslog(LOG_ERR, "connect: %m\n");
		exit(1);
	}
#ifdef ACCESS_CONTROL
	/*
	 * We now know the IP address of the client.  Find out its
	 * name for future logging or access restrictions.  Log any
	 * mismatches between the name and the address, and flag an
	 * absolute access failure (we won't talk to anyone trying to
	 * spoof us or who doesn't have a registered name).
	 */
	if (logging || restrict) {
	    haddr = inet_ntoa(from.sin_addr);
	    if (hent = gethostbyaddr(from.sin_addr, sizeof(from.sin_addr),
				     AF_INET)) {
		register char **aptr;
		
		host = malloc(strlen(hent->h_name) + 1);
		strcpy(host, hent->h_name);
		if (hent = gethostbyname(host)) {
		    for (aptr = hent->h_addr_list; *aptr; aptr++) {
			if (!bcmp(*aptr, from.sin_addr, hent->h_length))
			    break;
		    }
		}
		if ((hent == 0) || (*aptr == 0)) {
		    syslog(LOG_NOTICE,
			   "host address %s not found for host name %s",
			   haddr, host);
		    if (restrict) {
			access_fail++;
		    }
		}
	    }
	    else {
		syslog(LOG_NOTICE, "unknown host at address %s", haddr);
		host = haddr;
		if (restrict) {
		    access_fail++;
		}
	    }
	}
	/*
	 * If we have been requested to go secure, chroot() to the
	 * specified directory.
	 */
	if (secure) {
	    if (chroot(homedir) < 0) {
		syslog(LOG_ERR, "cannot chroot to directory %s: %m", homedir);
		exit(1);
	    }
	    chdir("/"); 	/* cd to the new root */
	}
	else {
	    chdir(homedir);
	}

	/*
	 * If we are running as root, it is because of one of two
	 * things:
	 *
	 *     1) we needed root privileges to chroot()
	 *     2) the sysadmin made a mistake
	 *
	 * Make things more secure by changing to the user "nobody"
	 */
	if (getuid() == 0) {
	    if (pw = getpwnam("nobody")) {
		uid = pw->pw_uid;
		gid = pw->pw_gid;
	    }
	    else {
		uid = UID_NOBODY;
		gid = GID_NOBODY;
	    }
	    if (setgid(gid) == -1) {
		syslog(LOG_ERR, "setgid(%d): %m", gid);
		exit(1);
	    }
	    if (setuid(uid) == -1) {
		syslog(LOG_ERR, "setuid(%d): %m", uid);
		exit(1);
	    }
	}
#endif
	tp = (struct tftphdr *)buf;
	tp->th_opcode = ntohs(tp->th_opcode);
	if (tp->th_opcode == RRQ || tp->th_opcode == WRQ)
		tftp(tp, n);
	exit(1);
}

int	validate_access();
int	sendfile(), recvfile();

struct formats {
	char	*f_mode;
	int	(*f_validate)();
	int	(*f_send)();
	int	(*f_recv)();
	int	f_convert;
} formats[] = {
	{ "netascii",	validate_access,	sendfile,	recvfile, 1 },
	{ "octet",	validate_access,	sendfile,	recvfile, 0 },
#ifdef notdef
	{ "mail",	validate_user,		sendmail,	recvmail, 1 },
#endif
	{ 0 }
};

/*
 * Handle initial connection protocol.
 */
tftp(tp, size)
	struct tftphdr *tp;
	int size;
{
	register char *cp;
	int first = 1, ecode;
	register struct formats *pf;
	char *filename, *mode;

	filename = cp = tp->th_stuff;
again:
	while (cp < buf + size) {
		if (*cp == '\0')
			break;
		cp++;
	}
	if (*cp != '\0') {
		nak(EBADOP);
		exit(1);
	}
	if (first) {
		mode = ++cp;
		first = 0;
		goto again;
	}
	for (cp = mode; *cp; cp++)
		if (isupper(*cp))
			*cp = tolower(*cp);
	for (pf = formats; pf->f_mode; pf++)
		if (strcmp(pf->f_mode, mode) == 0)
			break;
	if (pf->f_mode == 0) {
		nak(EBADOP);
		exit(1);
	}
	ecode = (*pf->f_validate)(filename, tp->th_opcode);
#ifdef ACCESS_CONTROL
	/*
	 * If we are logging, log who it is and what they wanted, and
	 * whether or not they succeeded.
	 */
	if (logging) {
	    syslog(LOG_NOTICE,
		   "connection from %s, file: %s, access mode: %s -- %s",
		   host, filename, tp->th_opcode == WRQ ? "write" : "read",
		   ecode ? "failed" : "succeeded");
	}
#endif
	if (ecode) {
#ifdef ACCESS_CONTROL
	    	/*
	     	 * A likely cause of an error here is that someone has
	    	 * broadcast a RRQ packet because he's being booted and
	    	 * doesn't know who his server is.  We delay sending an
	    	 * ERROR to allow time for the real server to respond (in
	    	 * case the client has a poor Ethernet connection).
	    	 */
	    	sleep(DELAY_SECS);
#endif
		nak(ecode);
		exit(1);
	}
	if (tp->th_opcode == WRQ)
		(*pf->f_recv)(pf);
	else
		(*pf->f_send)(pf);
	exit(0);
}


FILE *file;

/*
 * Validate file access.  Since we
 * have no uid or gid, for now require
 * file to exist and be publicly
 * readable/writable.
 * If we were invoked with arguments
 * from inetd then the file must also be
 * in one of the given directory prefixes.
 * Note also, full path name must be
 * given as we have no login directory.
 */
validate_access(filename, mode)
	char *filename;
	int mode;
{
	struct stat stbuf;
	int	fd;
#ifndef ACCESS_CONTROL
	char *cp, **dirp;

	if (*filename != '/')
		return (EACCESS);
	/*
	 * prevent tricksters from getting around the directory restrictions
	 */
	for (cp = filename + 1; *cp; cp++)
		if(*cp == '.' && strncmp(cp-1, "/../", 4) == 0)
			return(EACCESS);
	for (dirp = dirs; *dirp; dirp++)
		if (strncmp(filename, *dirp, strlen(*dirp)) == 0)
			break;
	if (*dirp==0 && dirp!=dirs)
		return (EACCESS);
#endif
	if (stat(filename, &stbuf) < 0)
		return (errno == ENOENT ? ENOTFOUND : EACCESS);
	if (mode == RRQ) {
		if ((stbuf.st_mode&(S_IREAD >> 6)) == 0)
			return (EACCESS);
	} else {
		if ((stbuf.st_mode&(S_IWRITE >> 6)) == 0)
			return (EACCESS);
	}
#ifdef ACCESS_CONTROL
	if ((stbuf.st_mode & S_IFMT) != S_IFREG)
	    return (EACCESS);
	if (check_access(filename) == 0)
	    return (EACCESS);
#endif
	fd = open(filename, mode == RRQ ? 0 : 1);
	if (fd < 0)
		return (errno + 100);
	file = fdopen(fd, (mode == RRQ)? "r":"w");
	if (file == NULL) {
		return errno+100;
	}
	return (0);
}

int	timeout;
jmp_buf	timeoutbuf;

void
timer()
{

	timeout += rexmtval;
	if (timeout >= maxtimeout)
		exit(1);
	longjmp(timeoutbuf, 1);
}

/*
 * Send the requested file.
 */
sendfile(pf)
	struct formats *pf;
{
	struct tftphdr *dp, *r_init();
	register struct tftphdr *ap;    /* ack packet */
	register int block = 1, size, n;

	signal(SIGALRM, timer);
	dp = r_init();
	ap = (struct tftphdr *)ackbuf;
	do {
		size = readit(file, &dp, pf->f_convert);
		if (size < 0) {
			nak(errno + 100);
			goto abort;
		}
		dp->th_opcode = htons((u_short)DATA);
		dp->th_block = htons((u_short)block);
		timeout = 0;
		(void) setjmp(timeoutbuf);

send_data:
		if (send(peer, dp, size + 4, 0) != size + 4) {
			syslog(LOG_ERR, "tftpd: write: %m\n");
			goto abort;
		}
		read_ahead(file, pf->f_convert);
		for ( ; ; ) {
			alarm(rexmtval);        /* read the ack */
			n = recv(peer, ackbuf, sizeof (ackbuf), 0);
			alarm(0);
			if (n < 0) {
				syslog(LOG_ERR, "tftpd: read: %m\n");
				goto abort;
			}
			ap->th_opcode = ntohs((u_short)ap->th_opcode);
			ap->th_block = ntohs((u_short)ap->th_block);

			if (ap->th_opcode == ERROR)
				goto abort;
			
			if (ap->th_opcode == ACK) {
				if (ap->th_block == block) {
					break;
				}
				/* Re-synchronize with the other side */
				(void) synchnet(peer);
				if (ap->th_block == (block -1)) {
					goto send_data;
				}
			}

		}
		block++;
	} while (size == SEGSIZE);
abort:
	(void) fclose(file);
}

void
justquit()
{
	exit(0);
}


/*
 * Receive a file.
 */
recvfile(pf)
	struct formats *pf;
{
	struct tftphdr *dp, *w_init();
	register struct tftphdr *ap;    /* ack buffer */
	register int block = 0, n, size;

	signal(SIGALRM, timer);
	dp = w_init();
	ap = (struct tftphdr *)ackbuf;
	do {
		timeout = 0;
		ap->th_opcode = htons((u_short)ACK);
		ap->th_block = htons((u_short)block);
		block++;
		(void) setjmp(timeoutbuf);
send_ack:
		if (send(peer, ackbuf, 4, 0) != 4) {
			syslog(LOG_ERR, "tftpd: write: %m\n");
			goto abort;
		}
		write_behind(file, pf->f_convert);
		for ( ; ; ) {
			alarm(rexmtval);
			n = recv(peer, dp, PKTSIZE, 0);
			alarm(0);
			if (n < 0) {            /* really? */
				syslog(LOG_ERR, "tftpd: read: %m\n");
				goto abort;
			}
			dp->th_opcode = ntohs((u_short)dp->th_opcode);
			dp->th_block = ntohs((u_short)dp->th_block);
			if (dp->th_opcode == ERROR)
				goto abort;
			if (dp->th_opcode == DATA) {
				if (dp->th_block == block) {
					break;   /* normal */
				}
				/* Re-synchronize with the other side */
				(void) synchnet(peer);
				if (dp->th_block == (block-1))
					goto send_ack;          /* rexmit */
			}
		}
		/*  size = write(file, dp->th_data, n - 4); */
		size = writeit(file, &dp, n - 4, pf->f_convert);
		if (size != (n-4)) {                    /* ahem */
			if (size < 0) nak(errno + 100);
			else nak(ENOSPACE);
			goto abort;
		}
	} while (size == SEGSIZE);
	write_behind(file, pf->f_convert);
	(void) fclose(file);            /* close data file */

	ap->th_opcode = htons((u_short)ACK);    /* send the "final" ack */
	ap->th_block = htons((u_short)(block));
	(void) send(peer, ackbuf, 4, 0);

	signal(SIGALRM, justquit);      /* just quit on timeout */
	alarm(rexmtval);
	n = recv(peer, buf, sizeof (buf), 0); /* normally times out and quits */
	alarm(0);
	if (n >= 4 &&                   /* if read some data */
	    dp->th_opcode == DATA &&    /* and got a data block */
	    block == dp->th_block) {	/* then my last ack was lost */
		(void) send(peer, ackbuf, 4, 0);     /* resend final ack */
	}
abort:
	return;
}

struct errmsg {
	int	e_code;
	char	*e_msg;
} errmsgs[] = {
	{ EUNDEF,	"Undefined error code" },
	{ ENOTFOUND,	"File not found" },
	{ EACCESS,	"Access violation" },
	{ ENOSPACE,	"Disk full or allocation exceeded" },
	{ EBADOP,	"Illegal TFTP operation" },
	{ EBADID,	"Unknown transfer ID" },
	{ EEXISTS,	"File already exists" },
	{ ENOUSER,	"No such user" },
	{ -1,		0 }
};

/*
 * Send a nak packet (error message).
 * Error code passed in is one of the
 * standard TFTP codes, or a UNIX errno
 * offset by 100.
 */
nak(error)
	int error;
{
	register struct tftphdr *tp;
	int length;
	register struct errmsg *pe;

	tp = (struct tftphdr *)buf;
	tp->th_opcode = htons((u_short)ERROR);
	tp->th_code = htons((u_short)error);
	for (pe = errmsgs; pe->e_code >= 0; pe++)
		if (pe->e_code == error)
			break;
	if (pe->e_code < 0) {
		pe->e_msg = strerror(error - 100);
		tp->th_code = EUNDEF;   /* set 'undef' errorcode */
	}
	strcpy(tp->th_msg, pe->e_msg);
	length = strlen(pe->e_msg);
	tp->th_msg[length] = '\0';
	length += 5;
	if (send(peer, buf, length, 0) != length)
		syslog(LOG_ERR, "nak: %m\n");
}

char err[40];

char *strerror(code)
     int code;
{
    if (code < sys_nerr)
	return(sys_errlist[code]);
    else {
	sprintf(err, "unknown error code %d", code);
	return(err);
    }
}

#ifdef ACCESS_CONTROL

#define SPACE	" \t\n"
#define MAXLINE	100

static char linebuf[MAXLINE];

/*
 * This is a utility routine to encapsulate the code to pull a dotted
 * quad, or a long hex number from the input line buffer.  It is used
 * by the read_config() routine.
 */
static unsigned long getIP(linebuf)
     register char *linebuf;
{
    register char *cp;
    register unsigned long a = 0;

    errno = 0;
    if (!(cp = strtok(linebuf, SPACE))) {
	errno = -1;
	return(a);
    }

    if ((int) (a = inet_addr(cp)) == -1) {	/* could be an error */
	if (strcmp(cp, "255.255.255.255") && strcmp(cp, "0xffffffff")) {
	    /* this was an error */
	    errno = EINVAL;
	}
    }
    return(a);
}

/*
 * This routine reads the configuration file (really an access control
 * file), parses the lines, and stores them in a linked list used
 * later for access checks.  It returns the linked list to its caller.
 * Any error causes it to log an error message an return a NULL list.
 * No cleanup is done since the caller will likely exit soon, anyway.
 */
static struct aentry *read_config(filename)
     char *filename;
{
    struct aentry *alist = NULL;
    register struct aentry *entry;
    register char *cp;
    FILE *conf;
    int line = 0;

    if (!filename) {
	syslog(LOG_ERR, "read_config(): no filename parameter?");
	return(NULL);
    }
    if (!(conf = fopen(filename, "r"))) {
	syslog(LOG_ERR, "unable to open %s: %m", filename);
	return(NULL);
    }

    while(fgets(linebuf, MAXLINE, conf)) {
	line++;
	if (linebuf[0] == '#')	/* skip comment lines */
	    continue;

	if (!(entry = (struct aentry *) malloc(sizeof(struct aentry)))) {
	    syslog(LOG_ERR, "out of memory at line %d of %s", line, filename);
	    alist=NULL;
	    break;
	}
	entry->next = alist;
	alist = entry;

	/* get the address field */
	entry->addr = getIP(linebuf);
	if (errno < 0)
	    goto format_error;
	if (errno == EINVAL) {
	    syslog(LOG_ERR, "invalid address on line %d of %s", line,
		   filename);
	    alist=NULL;
	    break;
	}

	/* get the mask field */
	entry->mask = getIP(NULL);
	if (errno < 0)
	    goto format_error;
	if (errno == EINVAL) {
	    syslog(LOG_ERR, "invalid mask on line %d of %s", line,
		   filename);
	    alist = NULL;
	    break;
	}

	/* get the pattern field */
	if (!(cp = strtok(NULL, SPACE)))
	    goto format_error;
	entry->pattern = strdup(cp);

	/* there shouldn't be anything else */
	if (strtok(NULL, SPACE) == NULL)
	    continue;

format_error:
	alist = NULL;
	syslog(LOG_ERR, "config format error on line %d of %s",
	       line, filename);
	break;
    }
    fclose(conf);
    return(alist);
}

/*
 * This routine scans the access list to determine if the requester
 * can access the file.  Regardless of the success of this test, the
 * file must still be world read/writeable.
 */
check_access(filename)
     register char *filename;
{
    register struct aentry *entry;

    if (access_fail)
	return(0);

    /*
     * NOTE: There is a flag to fnmatch (FNM_QUOTE) that would allow
     * the config file to contain "quoted" special characters (ie, you
     * could specify the pattern "a\*").  I don't know if I will need
     * it, but it can be added.
     */
    for (entry = access_list; entry; entry = entry->next) {
	if (((entry->mask & from.sin_addr.s_addr) == entry->addr) &&
	    fnmatch(entry->pattern, filename, 0)) {
	    break;
	}
    }

    return((int) entry);
}

usage(prog)
     char *prog;
{
    syslog(LOG_ERR, "usage: %s [-lrs] [homedir]", prog);
}
#endif
